#pfp-manage-form {
	position: relative;
	width: 100%;
}
.pfp-block { float: left; }

.pfp-button-left { float: left; }

/* PERM SELECT BLOCK */
#pfp-select-block {
	margin-left: 40px;
	float: left;
	max-width: 75%;
	min-width: 50%;
}

#pfp-select-form {
	margin-top: 40px;
}

#pfp-select-form table {
	overflow: auto;
	display: block;
}

#pfp-select-form td {
	white-space: nowrap;
}

.pagination {
	margin: 5px 0;
}

/* FILTER COLUMNS */

#pfp-perm-filter {
    z-index: 1000;
    width: 20%;
    background: #f8f8f8;
    border-left: none;
    margin: 0;
		float: left;
}

#pfp-perm-filter h2 {
    font-size: 14px;
    text-transform: uppercase;
    letter-spacing: 0.5px;
    padding: 5px 15px;
    margin-bottom: 12px;
    border-bottom: none;
}

#pfp-perm-filter h3 {
    font-weight: 400;
    font-size: 14px;
    padding: 0 15px;
    margin-bottom: 10px;
}

#pfp-perm-filter div {
    margin: 5px 0;
    padding: 0 15px 15px;
    border-bottom: 1px solid #eaeaea;
		overflow: auto;
}


#pfp-role-filter {
    z-index: 1000;
    background: #f8f8f8;
    border-left: none;
}

#pfp-role-filter h2 {
    font-size: 14px;
    text-transform: uppercase;
    letter-spacing: 0.5px;
    padding: 5px 15px;
    margin-bottom: 12px;
    border-bottom: none;
}

#pfp-perm-filter ul.errorlist, #pfp-role-filter ul.errorlist {
	background: none;
}

#pfp-role-filter li { list-style-type: none; }

